/*
	No More Studios

	May 6th, 2015
	
	Dungeon

	This program is a game called Dungeon. The objective is for the
	player to reach the end of each dungeon by destroying all monsters
	in each dungeon. Upon destroying all monsters, the player will advance
	to the next level. There are 3 levels in total, see if you can beat it!

	To see all options for commands, press ? and then press enter.

	Game is easy as the monsters don't move around, but is
	kinda neat still.

	Challenge:

	Try to program AI for the monsters. Try to make
	it so that the monsters move everytime the user
	either rests, attacks, or moves.

*/
#include <iostream>
#include <string>
#include <time.h>
#include <fstream>
using namespace std;

// Classes
class Monster { // Class is for each monster's blueprint.
	private:
		int maxHealth;
		int health, x, y, id;
	public:
		Monster() {
			health = 10;
			maxHealth = 10;
		}

		Monster(int health, int x, int y){
			health = 10;
			maxHealth = 10;

			this->x = x;
			this->y = y;
			this->health = health;
		}
		int getHealth() { return health; }
		int getX() { return x; }
		int getY() { return y; }
		int getMaxHealth() { return maxHealth; }
		void setHealth(int h) { health = h; }
		void setX(int x) { this->x = x; }
		void setY(int y) { this->y = y; }
};

// Prototypes
void drawUI();
void drawGameBoard();
void placeCharacter();
bool validInput(char command);
void displayHelpMenu();
bool validAction(char action);
void initMonsterAction();
void placeMonsters();
void displayMainMenu();
void rollDamage(int);
void saveGame();
bool loadGame();
void checkWin();
void nextLevel();

// Player global variables
int playerX = 0; // These two variables to easily figure out
int playerY = 0; // where the player is in the map array.
char playerIcon = 'H';
int const MAX_HEALTH = 10;
int playerHealth = 10;

// Monster global variables
char monsterIcon = 'm';

// Map global variables
int const MAP_WIDTH = 12;
int const MAP_HEIGHT = 12;
string gameMap[3][MAP_HEIGHT][MAP_WIDTH] = // levels
{
	{
		{"0","0","0","0","0","0","0","0","0","0","0","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0","0","0","0","0","0","0","0","0","0","0","0" }
	},
	{
		{"0","0","0","0","0","0","0","0","0","0","0","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," ","0"," "," ","0" },
		{"0"," ","0"," "," "," "," "," ","0"," "," ","0" },
		{"0"," ","0"," "," "," "," "," "," "," "," ","0" },
		{"0"," ","0"," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," ","0"," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0","0","0","0","0","0","0","0","0","0","0","0" }
	},
	{
		{"0","0","0","0","0","0","0","0","0","0","0","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," "," "," "," "," "," "," "," "," "," ","0" },
		{"0"," ","0"," "," ","0"," "," "," ","0"," ","0" },
		{"0"," ","0"," "," "," "," "," "," "," "," ","0" },
		{"0"," ","0"," "," "," "," ","0"," "," "," ","0" },
		{"0"," ","0"," "," "," "," ","0"," "," "," ","0" },
		{"0"," "," "," ","0"," "," "," "," "," "," ","0" },
		{"0"," "," "," ","0"," ","0"," ","0","0","0","0" },
		{"0"," "," "," "," "," ","0"," "," "," "," ","0" },
		{"0"," "," "," "," "," ","0"," "," "," "," ","0" },
		{"0","0","0","0","0","0","0","0","0","0","0","0" }
	}
};

// Game mechanics variables
bool inBattleMode = false;
int monsterInBattle = 0;

int currentLevel = 0;
Monster monsterArray[5];

int main() {

	displayMainMenu();
	placeMonsters();
	placeCharacter();

	while(true) { // simple game loop
		char command = ' ';
		
		drawGameBoard();
		cout << "\n";

		while(!validInput(command)) { // Main user validation loop
			cout << "Enter command now. To see the list of options, type ? and enter: ";
			cin >> command;
			cin.ignore(10000, '\n');
		}
		checkWin();
	}
	
	return 0;
}

void displayMainMenu() { // Main Menu
	cout << "000  0  0 0  0 0000 000 0000 0  0" << endl;
	cout << "0  0 0  0 00 0 0    0   0  0 00 0" << endl;
	cout << "0  0 0  0 0 00 0 00 00  0  0 0 00" << endl;
	cout << "0  0 0  0 0  0 0  0 0   0  0 0  0" << endl;
	cout << "000  0000 0  0 0000 000 0000 0  0" << endl;
	cout << "\nWelcome to Dungeon, a dungeon crawler game created by Bradley Leopold.";
	cout << "\n\nIn this game, the objective is to clear out each dungeon of monsters. Upon doing so, you will advance to the next level.";
	cout << "\n\n";
	system("pause");
}

void drawGameBoard() { // Basically just draws the levels on screen, as well as the UI each time its called.
	system("cls");
	drawUI();

	for(int i = 0; i < MAP_HEIGHT; i++) {
		for(int j = 0; j < MAP_WIDTH; j++) {
			if(gameMap[currentLevel][i][j] == "0") {
				cout << char(219);
			} else {
				cout << gameMap[currentLevel][i][j];
			}
			if(i == 0 || i == MAP_HEIGHT-1) {
				if(j < MAP_WIDTH - 1) {
					cout << char(219);
				}
			} else {
				cout << " ";
			}
		}
		cout << "\n";
	}
}

void drawUI() {
	cout << "Health: ";
	for(int i = 0; i < MAX_HEALTH; i++) {
		if(i < playerHealth) {
			cout << (char) 219; //ascii for the solid block
		} else {
			cout << (char) 176; //ascii for the spotted block
		}
	}

	if(inBattleMode) {
		cout << " Monster Health: ";
		for(int i = 0; i < MAX_HEALTH; i++) {
			if(i < monsterArray[monsterInBattle].getHealth()) {
				cout << (char) 219;
			} else {
				cout << (char) 176;
			}
		}
	}
	cout << "\n\n";
}

void placeCharacter() { // places character on the map as well as the player's coordinates
	if(currentLevel == 0) {
		playerX = 1;
		playerY = 1;
	} else if(currentLevel == 1) {
		playerX = 5;
		playerY = 1;
	} else if(currentLevel == 2) {
		playerX = 1;
		playerY = 7;
	}
	for(int i = 0; i < MAP_HEIGHT; i++) {
		for(int j = 0; j < MAP_WIDTH; j++) {
			if(gameMap[currentLevel][i][j] == "H") {
				gameMap[currentLevel][i][j] = " ";
			}
		}
	}
	if(gameMap[currentLevel][playerY][playerX] != "H") {
		gameMap[currentLevel][playerY][playerX] = "H";
	}
}

void placeMonsters() {
	if(currentLevel == 0) {

		monsterArray[0] = Monster(10, 2, 2);
		monsterArray[1] = Monster(10, 2, 9);

		gameMap[currentLevel][2][2] = monsterIcon;
		gameMap[currentLevel][9][2] = monsterIcon;
	} else if(currentLevel == 1) {
		monsterArray[0] = Monster(10, 1, 1);
		monsterArray[1] = Monster(10, 6, 6);
		monsterArray[2] = Monster(10, 10, 10);
		gameMap[currentLevel][1][1] = monsterIcon;
		gameMap[currentLevel][6][6] = monsterIcon;
		gameMap[currentLevel][10][10] = monsterIcon;
	} else if(currentLevel == 2) {
		monsterArray[0] = Monster(10, 2, 1);
		monsterArray[1] = Monster(10, 5, 4);
		monsterArray[2] = Monster(10, 5, 5);
		monsterArray[3] = Monster(10, 10, 6);
		monsterArray[4] = Monster(10, 2, 10);
		gameMap[currentLevel][1][2] = monsterIcon;
		gameMap[currentLevel][4][5] = monsterIcon;
		gameMap[currentLevel][5][5] = monsterIcon;
		gameMap[currentLevel][6][10] = monsterIcon;
		gameMap[currentLevel][10][2] = monsterIcon;
	}
}

bool validInput(char command) { // Massive function to determine what the user wants to do

	if(command == 'w' || command == 'W') {
		if(gameMap[currentLevel][playerY-1][playerX] == " ") {
			gameMap[currentLevel][playerY][playerX] = " ";
			gameMap[currentLevel][playerY-1][playerX] = playerIcon;
			playerY -= 1;
			return true;
		} else if(gameMap[currentLevel][playerY-1][playerX] == "m"){
			initMonsterAction();
			for(int i = 0; i < sizeof(monsterArray) / sizeof(monsterArray[0]); i++) {
				if(monsterArray[i].getX() == playerX) {
					if((monsterArray[i].getY()+1) == playerY) {
						rollDamage(i);
						inBattleMode = true;
						monsterInBattle = i;
					}
				}
			}
			return true;
		} else {
			return false;
		}
	} else if(command == 'a' || command == 'A') {
		if(gameMap[currentLevel][playerY][playerX-1] == " ") {
			gameMap[currentLevel][playerY][playerX] = " ";
			gameMap[currentLevel][playerY][playerX-1] = playerIcon;
			playerX -= 1;
			return true;
		} else if(gameMap[currentLevel][playerY][playerX-1] == "m"){
			initMonsterAction();
			for(int i = 0; i < sizeof(monsterArray) / sizeof(monsterArray[0]); i++) {
				if(monsterArray[i].getX()+1 == playerX) {
					if((monsterArray[i].getY()) == playerY) {
						rollDamage(i);
						inBattleMode = true;
						monsterInBattle = i;
					}
				}
			}
			return true;
		}else {
			return false;
		}
	} else if(command == 's' || command == 'S') {
		if(gameMap[currentLevel][playerY+1][playerX] == " ") {
			gameMap[currentLevel][playerY][playerX] = " ";
			gameMap[currentLevel][playerY+1][playerX] = playerIcon;
			playerY += 1;
			return true;
		} else if(gameMap[currentLevel][playerY+1][playerX] == "m"){
			initMonsterAction();
			for(int i = 0; i < sizeof(monsterArray) / sizeof(monsterArray[0]); i++) {
				if(monsterArray[i].getX() == playerX) {
					if((monsterArray[i].getY()-1) == playerY) {
						rollDamage(i);
						inBattleMode = true;
						monsterInBattle = i;
					}
				}
			}
			return true;
		} else {
			return false;
		}
	} else if(command == 'd' || command == 'D') {
		if(gameMap[currentLevel][playerY][playerX+1] == " ") {
			gameMap[currentLevel][playerY][playerX] = " ";
			gameMap[currentLevel][playerY][playerX+1] = playerIcon;
			playerX += 1;
			return true;
		} else if(gameMap[currentLevel][playerY][playerX+1] == "m"){
			initMonsterAction();
			for(int i = 0; i < sizeof(monsterArray) / sizeof(monsterArray[0]); i++) {
				if(monsterArray[i].getX()-1 == playerX) {
					if((monsterArray[i].getY()) == playerY) {
						rollDamage(i);
						inBattleMode = true;
						monsterInBattle = i;
					}
				}
			}
			return true;
		} else {
			return false;
		}
	} else if(command == 'r') {
		if(playerHealth < MAX_HEALTH) {
			playerHealth += 1;
		}
		return true;
	} else if(command == '?') {
		displayHelpMenu();
		return false;
	} else if(command == 'q') {
		saveGame();
		cout << "Game Saved\n";
		return false;
	} else if(command == 'l') {
		if(loadGame()) {
			placeCharacter();
			placeMonsters();
			playerHealth = 10;
			cout << "Game Loaded";
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

bool validAction(char action) {
	if(action == 'a') {
		return true;
	} else if(action == 'p') {
		return true;
	} else if(action == 'r') {
		return true;
	} else {
		return false;
	}
}

void rollDamage(int i) {
	monsterArray[i].setHealth(monsterArray[i].getHealth() - 2);
	if(monsterArray[i].getHealth() <= 0) {
		gameMap[currentLevel][monsterArray[i].getY()][monsterArray[i].getX()] = " ";
	}
}

void displayHelpMenu() {
	cout << "Help Menu\n";
	cout << "-----------\n";
	cout << "The objective of this game is to clear out each dungeon of monsters.\n";
	cout << "To attack monsters, walk to them completely. Once you walk in the direction of\n";
	cout << "that monster, you will then deal damage to it. Get its health to 0 to defeat it.\n\n";
	cout << "Once all monsters are defeated in the dungeon, you will go to the next level.\n\n";

	// Map legend
	cout << "Map Icon Legend\n";
	cout << "-----------\n";
	cout << "Player's Icon: " << playerIcon << "\n";
	cout << "Monster's Icon: " << monsterIcon << "\n\n";

	// commands
	cout << "Commands you can perform\n";
	cout << "-----------\n";
	cout << "w, a, s, and d to move character, attack monster in that direction.\n";
	cout << "r: Makes character rest, healing one hitpoint. Uses a turn.\n";
	cout << "q: Saves which level you are on. Does not save anything else.\n";
	cout << "l: Loads game save file.\n";
}

void initMonsterAction() { // rolls the monsters attack
	srand(time(NULL));
	int monsterRoll = rand() % 10;	
	
	if(monsterRoll < 7) {
		playerHealth -= 1;
	}
}

void checkWin() { // checks every loop if the user has won the game, then advances the level.
	int counter = 0;

	if(playerHealth > 0) {
		for(int i = 0; i < MAP_HEIGHT; i++) {
			for(int j = 0; j < MAP_WIDTH; j++) {
				if(gameMap[currentLevel][i][j] == "m") {
					counter += 1;
				}
			}
		}
		if(counter <= 0) {
			if(currentLevel < 2) {
				system("cls");
				cout << "Congrats on beating level " << currentLevel+1 << "!\n";
				cout << "Press enter to proceed to the next level...";
			} else {
				system("cls");
				cout << "You have beaten the game!!!!! Congrats on your victory!!!!\n";
				cout << "You may exit the game, over press enter to start at the beginning.\n";
			}
			cin.clear();
			cin.ignore(10000, '\n');
			nextLevel();
		}
	} else {
		system("cls");
		cout << "Sorry, your character has died\n";
		cout << "Press enter to restart at the beginning of the game...";
		playerHealth = 10;
		currentLevel = 0;
		placeCharacter();
		placeMonsters();
		cin.clear();
		cin.ignore(10000, '\n');
	}
}

void nextLevel() { // changes level, resets some values
	if(currentLevel < 2) {
		currentLevel += 1;
	} else {
		currentLevel = 0;
		playerHealth = 10;
	}
	inBattleMode = false;
	monsterInBattle = 0;
	placeCharacter();
	placeMonsters();
}

void saveGame() { // Save game file, only saves which level the player was on.
	ofstream file;
	file.open("gameSave.txt");
	file << currentLevel;
	file.close();
}

bool loadGame() { // Load game saved level
	ifstream file("gameSave.txt");
	if(!file) {
		cout << "Sorry, but no game saves were detected.\n";
		return false;
	} else {
		while(!file.eof()) {
			file >> currentLevel;
		}
		gameMap[currentLevel][playerY][playerX] = " ";
		file.close();
		return true;
	}
}